#include <stdio.h>
#include <sys/times.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>

//-- timer for benchmarks ------------------------------------------------------

double
walltime()
{
   struct tms    ts;
   static double ClockTick=0.0;

   if (ClockTick==0.0) {
        ClockTick = 1.0 / ((double) sysconf(_SC_CLK_TCK));
   }
   return ((double) times(&ts)) * ClockTick;
}

//-- setup and print matrices --------------------------------------------------

void
initGeMatrix(int m, int n, double *A, int incRowA, int incColA)
{
    int i, j;

    for (i=0; i<m; ++i) {
        for (j=0; j<n; ++j) {
            A[i*incRowA+j*incColA] = i*n + j + 1;
        }
    }
}

void
randGeMatrix(int m, int n, double *A, int incRowA, int incColA)
{
    int i, j;
    for (j=0; j<n; ++j) {
        for (i=0; i<m; ++i) {
            A[i*incRowA+j*incColA] = ((double)rand()-RAND_MAX/2)*200/RAND_MAX;
        }
    }
}

void
makeTrlDiagDom(int n, int unitDiag, double *A, int incRowA, int incColA)
{
    int i, j;

    for (j=0; j<n; ++j) {
        double asum = 0;
        double A_jj = (unitDiag) ? 1
                                 : A[j*(incRowA+incColA)];
        for (i=j+1; i<n; ++i) {
            asum += fabs(A[i*incRowA+j*incColA]);
        }
        for (i=j+1; i<n; ++i) {
            A[i*incRowA+j*incColA] *= A_jj/(asum*1.1);
        }
    }
}

void
makeGeDiagDom(int n, int m, double *A, int incRowA, int incColA)
{
    int i, j;

    for (j=0; j<n; ++j) {
        double asum = 0;

        for (i=0; i<m; ++i) {
            if (i!=j) {
                asum += fabs(A[i*incRowA+j*incColA]);
            }
        }
        for (i=0; i<m; ++i) {
            if (i!=j) {
                A[i*incRowA+j*incColA] *= A[j*(incRowA+incColA)]/(asum*1.1);
            }
        }
    }
}

void
printGeMatrix(int m, int n, const double *A, int incRowA, int incColA)
{
    int i, j;

    for (i=0; i<m; ++i) {
        for (j=0; j<n; ++j) {
            printf("%10.4lf ", A[i*incRowA+j*incColA]);
        }
        printf("\n");
    }
    printf("\n\n");
}

void
printTrMatrix(int m, int n, int unitDiag, int lower,
              const double *A, int incRowA, int incColA)
{
    int i, j;

    for (i=0; i<m; ++i) {
        for (j=0; j<n; ++j) {
            if (unitDiag && (i==j)) {
                printf("%10.4lf ", 1.0);
            } else if ((lower && (i>=j)) || (!lower && (i<=j))) {
                printf("%10.4lf ", A[i*incRowA+j*incColA]);
            } else {
                printf("%10.4lf ", 0.0);
            }
        }
        printf("\n");
    }
    printf("\n\n");
}
